uniform float windWaveCounter;
uniform float windWaveCounterHighFreq;
uniform float waterWaveCounter;
uniform float windSpeed;
uniform vec3 playerpos;
uniform float globalWarpIntensity;

uniform float glitchStrength = 0;
uniform float windWaveIntensity = 1;

const vec3 vn1 = vec3(0.0,0.0,1.0);
const vec3 vn2 = vec3(0.0,1.0,0.0);
const vec3 vn3 = vec3(0.0,1.0,1.0);
const vec3 vn4 = vec3(1.0,0.0,0.0);
const vec3 vn5 = vec3(1.0,0.0,1.0);
const vec3 vn6 = vec3(1.0,1.0,0.0);
const vec3 vn7 = vec3(1.0,1.0,1.0);

//const mat3 m = mat3( 127.1, 311.7, 74.7,
//                    269.5, 183.3, 246.1,
//                    113.5, 271.9, 124.6 );


// Value noise by https://www.shadertoy.com/view/4sfGzS; 30.6.21 radfast updated with part of the mod289 approach from https://gist.github.com/patriciogonzalezvivo/670c22f3966e662d2f83
vec3 ghash( vec3 p )
{
	vec3 o;
	// these constants are the matrix m
	// individual components multiplied, because the whole matrix multiplication produces float rounding differences from C# equivalent code (Bell pepper)
	o.x = 127.1 * p.x + 311.7 * p.y + 74.7 * p.z;
	o.y = 269.5 * p.x + 183.3 * p.y + 246.1 * p.z;
	o.z = 113.5 * p.x + 271.9 * p.y + 124.6 * p.z;
	vec3 q = ((o * 0.025) + 8.0) * o;   // the constants 4.25 and 8.0 found empirically to give similar noise distribution to the sin approach
	return -1.0 + 2.0*fract(mod(q, 289.0) * (1.0 / 41.0));
}

float gnoise( in vec3 p )
{
    vec3 i = floor( p );
    vec3 f = p - i;
	
    vec3 u = f*f*(3.0-2.0*f);

    vec4 a = vec4 ( dot(ghash(i), f),
                    dot(ghash(i + vn1), f - vn1),
    		    dot(ghash(i + vn2), f - vn2),
                    dot(ghash(i + vn3), f - vn3));
    vec4 b = vec4 ( dot(ghash(i + vn4), f - vn4),
                    dot(ghash(i + vn5), f - vn5),
    		    dot(ghash(i + vn6), f - vn6),
                    dot(ghash(i + vn7), f - vn7));
    
    vec4 c = mix(a, b, u.x);
    vec2 rg = mix(c.xy, c.zw, u.y);

    // Added 1.2 here because our old noise was stronger
    return 1.2 * mix(rg.x, rg.y, u.z);
}

vec4 applyLiquidWarping(bool windAffected, vec4 worldPos, float div) {
	#if WAVINGSTUFF == 1
	vec3 noisepos = vec3((worldPos.x + playerpos.x) / 3, (worldPos.z + playerpos.z) / 3, waterWaveCounter / 8 + (windAffected ? windWaveCounter / 6 : 0));
	worldPos.y += gnoise(noisepos) / div;
	#endif
	
	return worldPos;
}

vec4 applyVertexWarping(int renderFlags, vec4 worldPos) {

	#if WAVINGSTUFF == 1

	// bit 11 is the grass swaying on/off flag, 
	// bit 12 is water
	//	bit 27 is leaves windwave, any of them set to 1 needs to apply a transform, can test all these flags together in a single conditional -> faster exit for the majority of blocks which have no flag set

	if ((renderFlags & 0x8001800) > 0)
	{
		// bit 12 is the water waving on/off flag
		if ((renderFlags & 0x1000) > 0) {
			vec3 noisepos = vec3((worldPos.x + playerpos.x) / 3, (worldPos.z + playerpos.z) / 3, waterWaveCounter / 8 + windWaveCounter / 6);
			worldPos.y += gnoise(noisepos) / 10;

		} else {
			// Coopers reed water layer has bit 11 and 12 set. Lets just ignore bit 11 in this case
			
			float div = 30;            

			float x = worldPos.x + playerpos.x;
			float y = worldPos.y + playerpos.y;
			float z = worldPos.z + playerpos.z;
				
			float heightBend;
			float strength = windWaveIntensity * (1 + windSpeed);
			float counter;
			int windwaveConfig = 0;
			
			if ((renderFlags & 0x8000000) > 0)    //leaves
			{
				int groundDistance = (renderFlags >> 28) & 0x7;
				
				strength *=  ((renderFlags & 0x0800) > 0 ? 0 : 0.5);
				heightBend = groundDistance / 10.0;
				counter = windWaveCounterHighFreq;
			}
			else
			{
				windwaveConfig = (renderFlags >> 28) & 0x7;
				
				// windwaveConfig = 0  Default 
				// windwaveConfig = 1  On weak wave, also have only low frequency jiggle
				// windwaveConfig = 2  unused
				// windwaveConfig = 3  Solid fruit and Stalk, rotate with origin
				// windwaveConfig = 4  Fruit underleaves
				
				heightBend = 0.7;
				counter = windWaveCounter;

				if (windwaveConfig == 2) {
					heightBend = 0;
				} 
			
				if (windwaveConfig != 0)
				{
					if (windwaveConfig == 3)  //special for pineapples (stiff central stem, fruit must not distort)
					{
						x = floor(x) + 0.5;
						z = floor(z) + 0.5;
						y = floor(y) + 3 / 8.0;
						div *= 16.0 / 3;
						strength *= 3 / 16.0;
						heightBend = 3 / 4.0;
						x *= 5;  // for less uniformity across fields of pineapples
						z *= 5;  // for less uniformity across fields of pineapples
					}
					else if (windwaveConfig == 1)   //for stiffer plants, lower frequency and wave strength affected by vertex height on model
					{
						float heightFactor = (0.5 + y - floor(y));
						heightBend *= heightFactor;
						strength *= heightFactor;
						x *= 5;  // for less uniformity across fields of pineapples
						z *= 5;  // for less uniformity across fields of pineapples
					}
					else if (windwaveConfig == 4)   //keep x and z in line with 1 and 3
					{
						x *= 5;  // for less uniformity across fields of pineapples
						z *= 5;  // for less uniformity across fields of pineapples
					}
				}
			}
			
			float bendNoise = windSpeed * 0.2 + 1.4 * gnoise(vec3(mod(x, 4096.0) / 10, mod(z, 4096.0) / 10, mod(counter, 1024.0) / 4));

			float bend = windSpeed * (0.8 + bendNoise) * heightBend * windWaveIntensity;
			bend = min(4, bend);
			
			float wwaveHighFreq = windWaveCounterHighFreq;
			
			// weakwave
			if ((renderFlags & 0x4000) > 0) {
				bend *= ((renderFlags & 0x8000000) == 0) ? 0.2857143 : 0.5;
				div *= 3.5;
				if (windwaveConfig == 1 || windwaveConfig == 4) { wwaveHighFreq *= 0.5; strength *= 0.5; }
			}

			x += wwaveHighFreq;
			y += wwaveHighFreq;
			z += wwaveHighFreq;
			
			// Curve visualized: http://fooplot.com/#W3sidHlwZSI6MCwiZXEiOiIyKnNpbih4LzgpK3Npbih4LzIpK3NpbigwLjUrMip4KStzaW4oMSszKngpIiwiY29sb3IiOiIjMDAwMDAwIn0seyJ0eXBlIjoxMDAwLCJ3aW5kb3ciOlsiLTI0Ljc5NTUzMjIyNjU2MjQ4NiIsIjI0Ljc5NTUzMjIyNjU2MjQ4NiIsIi0xNS4yNTg3ODkwNjI0OTk5OTEiLCIxNS4yNTg3ODkwNjI0OTk5OTEiXX1d

			float dx = bend + strength * (sin(x*10)/120 + (2 * sin(x/2) + sin(x + y) + sin(0.5 + 4*x + 2*y) + sin(1 + 6*x + 3*y)/3) / div);
			float dz = strength * ((2 * sin(z/4) + sin(z + 3 * y) + sin(0.5 + 4*z + 2*y) + sin(1 + 6*z + y)/3) / div);
			
			if (windwaveConfig <= 2)
			{
				worldPos.x += dx;
				worldPos.y += strength * (sin(5*y)/15 + cos(10*x) / 10 + sin(3*z)/2 + cos(x*2)/2.2) / div;
				worldPos.z += dz;
			} else
			if (windwaveConfig == 3 || windwaveConfig == 4)
			{
				// For pineapples
				dx -= bend/4;
				float xx = worldPos.x + playerpos.x;
				float yy = worldPos.y + playerpos.y;
				float zz = worldPos.z + playerpos.z;
				xx = xx - floor(xx) - 0.5;
				yy = yy - floor(yy);
				if (yy < 0.25) yy += 1.0;
				zz = zz - floor(zz) - 0.5;
				float xangle = atan(dx, 3/4.0) / 4;
				float zangle = atan(dz, 3/4.0) / 4;
				float sinx = sin(xangle);
				float sinz = sin(zangle);
				float oneminuscosx = 1.0 - cos(xangle);
				float oneminuscosz = 1.0 - cos(zangle);
				worldPos.x += dx / 4 + yy * sinx - xx * oneminuscosx;
				worldPos.y += -xx * sinx - zz * sinz - yy * oneminuscosx - yy * oneminuscosz;
				worldPos.z += dz / 4 + yy * sinz - zz * oneminuscosz;
			}
		}
	}

	
	if (glitchStrength > 0.1) {
		float str = max(0, glitchStrength - 0.1);
		str *= clamp(length(worldPos) * glitchStrength - 1, 0, 150);
		
		float xf = (worldPos.x + playerpos.x) / 10;
		float zf = (worldPos.z + playerpos.z) / 10;
		worldPos.x += str * gnoise(vec3(xf, zf, windWaveCounter/6)) / 5;
		worldPos.y += str * gnoise(vec3(xf, zf, windWaveCounter/10)) / 5;
		worldPos.z += str * gnoise(vec3(xf, zf, windWaveCounter/3.5)) / 5;
	}
	
	if (globalWarpIntensity > 0) {
		float x = max(0, (mod(20*windWaveCounter, 30)) + (worldPos.x + playerpos.x) / 5.0 + (worldPos.y + playerpos.y) / 8.0 - 40);
		worldPos.x += (sin(x / 2) + sin(0.5 + 2*x) + sin(1 + 3*x)/3) / 30.0 * globalWarpIntensity;
		
		worldPos.z += (cos(x / 3) + cos(0.2 + 2.2*x) + cos(1 + 4*x)/3) / 30.0 * globalWarpIntensity;
	}
	
	#endif
		
	return worldPos;
}




